==> hperf-6.5.4.tar.gz: hperf/pipe.c <==
/*
    This file is part of HPerf.
    Copyright (C) 2020-2023  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "message.h"
#include "pipe.h"

// ************************************************************************
// 
// ************************************************************************
int recursive_mkdir(char *path_orig, int flags)
{
	char path[1024];
	
	strncpy(path, path_orig, sizeof(path));
	path[sizeof(path) - 1] = 0;
	
	char *p = path;
	mode_t mode = (flags & MKDIR_UNIVERSAL) ? 0777 : 0700;
	
	while (1) {
		char *slash = strchr(p, '/');
		
		if (slash) {
			if (slash == p) {
				p++;
				continue;
			}
			
			*slash = 0;
			p = slash + 1;
		} else {
			if (flags & MKDIR_PARENTS)
				return 0;
			if (*p == 0)
				return 0;
			p = NULL;
		}
		
		if (mkdir(path, mode)) {
			if (errno != EEXIST) {
				ERROR("%s: %s: %s\n",
					path_orig, path,
					strerror(errno));
				return -1;
			}
		}
		
		if (slash)
			*slash = '/';
		
		if (!p)
			return 0;
	}
	
}


// ************************************************************************
// 
// ************************************************************************
static int fd_move(int dst, int src)
{
	if (dup2(src, dst) != dst) {
		ERROR("dup2(): %s\n", strerror(errno));
		return -1;
	}
	
	if (src != dst)
		close(src);
	
	return 0;
}

static int fd_null(int fd)
{
	int mode = (fd == 0) ? O_RDONLY :
		((fd <= 2) ? O_WRONLY : O_RDWR);
	
	int fn = open("/dev/null", mode);
	
	if (fn < 0) {
		ERROR("open(\"/dev/null\"): %s\n", strerror(errno));
		return -1;
	}
	
	if (dup2(fn, fd) != fd) {
		ERROR("dup2(): %s\n", strerror(errno));
		return -1;
	}
	
	if (fn != fd)
		close(fn);
	
	return 0;
}

// ************************************************************************
// 
// ************************************************************************
#define TEE_BUFFER	((size_t)2 << 20)

int pipe_in(char **argv, char *tee_path)
{
	// open
	int tee_fd = -1;
	
	if (tee_path) {
		int mkdir_flags = MKDIR_PARENTS;
		
		if (tee_path[0] == '/')
			mkdir_flags |= MKDIR_UNIVERSAL;
		
		if (recursive_mkdir(tee_path, mkdir_flags))
			return -1;
		
		tee_fd = creat(tee_path, 0600);
		
		if (tee_fd < 0) {
			ERROR("%s: %s\n", tee_path, strerror(errno));
			return -1;
		}
	}
	
	// pipe
	int fd[2];
	
	if (pipe(fd)) {
		ERROR("pipe(): %s\n", strerror(errno));
		return -1;
	}
	
	// fork
	pid_t child = fork();
	
	if (child == (pid_t)-1) {
		ERROR("fork(): %s\n", strerror(errno));
		return -1;
	}
	
	// parent
	if (child != 0) {
		// close tee file
		if (tee_fd >= 0)
			close(tee_fd);
		
		// close pipe's write end
		close(fd[1]);
		
		return fd[0];
	}
	
	// child
	close(fd[0]);
	
	if (tee_fd >= 0) {
		// pipe
		int xfd[2];
		
		if (pipe(xfd)) {
			ERROR("pipe(): %s\n", strerror(errno));
			exit(1);
		}
		
		// fork
		pid_t xchild = fork();
		
		if (xchild == (pid_t)-1) {
			ERROR("fork(): %s\n", strerror(errno));
			exit(1);
		}
		
		// parent
		if (xchild != 0) {
			// close pipe's write end
			close(xfd[1]);
			
			char *buff = malloc(TEE_BUFFER);
			if (!buff) {
				ERROR("malloc() %s\n", strerror(errno));
				exit(1);
			}
			
			while (1) {
				ssize_t n = read(xfd[0], buff, TEE_BUFFER);
				
				if (n < 0) {
					ERROR("read(): %s\n",
						strerror(errno));
					exit(1);
				}
				
				if (n == 0)
					break;
				
				ssize_t w = write(fd[1], buff, n);
				
				if (w != n) {
					ERROR("write(%zd / %zd): %s\n",
						w, n, strerror(errno));
					exit(1);
				}
				
				ssize_t t = write(tee_fd, buff, n);
				
				if (t != n) {
					ERROR("%s: %zd / %zd: %s\n",
						tee_path,
						t, n, strerror(errno));
					exit(1);
				}
			}
			
			free(buff);
			close(xfd[0]);
			close(fd[1]);
			close(tee_fd);
			exit(0);
		}
		
		// child
		close(xfd[0]);
		close(fd[1]);
		close(tee_fd);
		
		fd[1] = xfd[1];
	}
	
	// exec
	if (fd_move(1, fd[1]))
		exit(1);
	
	if (verbosity < 2) {
		if (fd_null(2))
			exit(1);
	}
	
	execvp(argv[0], argv);
	
	ERROR("execvp('%s'): %s\n", argv[0], strerror(errno));
	exit(1);
}



// ************************************************************************
// 
// ************************************************************************

==> hperf-6.8.8.tar.gz: hperf/pipe.c <==
/*
    This file is part of HPerf.
    Copyright (C) 2020-2023  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "message.h"
#include "pipe.h"

// ************************************************************************
// 
// ************************************************************************
int file_readable(char *path)
{
	int fd = open(path, O_RDONLY);
	
	if (fd < 0)
		return 0;
	
	close(fd);
	
	return 1;
}

// ************************************************************************
// 
// ************************************************************************
int recursive_mkdir(char *path_orig, int flags)
{
	char path[1024];
	
	strncpy(path, path_orig, sizeof(path));
	path[sizeof(path) - 1] = 0;
	
	char *p = path;
	mode_t mode = (flags & MKDIR_UNIVERSAL) ? 0777 : 0700;
	
	while (1) {
		char *slash = strchr(p, '/');
		
		if (slash) {
			if (slash == p) {
				p++;
				continue;
			}
			
			*slash = 0;
			p = slash + 1;
		} else {
			if (flags & MKDIR_PARENTS)
				return 0;
			if (*p == 0)
				return 0;
			p = NULL;
		}
		
		if (mkdir(path, mode)) {
			if (errno != EEXIST) {
				ERROR("%s: %s: %s\n",
					path_orig, path,
					strerror(errno));
				return -1;
			}
		}
		
		if (slash)
			*slash = '/';
		
		if (!p)
			return 0;
	}
	
}


// ************************************************************************
// 
// ************************************************************************
static int fd_move(int dst, int src)
{
	if (dup2(src, dst) != dst) {
		ERROR("dup2(): %s\n", strerror(errno));
		return -1;
	}
	
	if (src != dst)
		close(src);
	
	return 0;
}

static int fd_null(int fd)
{
	int mode = (fd == 0) ? O_RDONLY :
		((fd <= 2) ? O_WRONLY : O_RDWR);
	
	int fn = open("/dev/null", mode);
	
	if (fn < 0) {
		ERROR("open(\"/dev/null\"): %s\n", strerror(errno));
		return -1;
	}
	
	if (dup2(fn, fd) != fd) {
		ERROR("dup2(): %s\n", strerror(errno));
		return -1;
	}
	
	if (fn != fd)
		close(fn);
	
	return 0;
}

// we cannot use exit() because it may flush stdio streams
#define child_exit	_exit

// ************************************************************************
// 
// ************************************************************************
#define TEE_BUFFER	((size_t)2 << 20)

int pipe_in(struct piped_input *p, char **argv, char *tee_path)
{
	// open
	int tee_fd = -1;
	
	if (tee_path) {
		int mkdir_flags = MKDIR_PARENTS;
		
		if (tee_path[0] == '/')
			mkdir_flags |= MKDIR_UNIVERSAL;
		
		if (recursive_mkdir(tee_path, mkdir_flags))
			return -1;
		
		tee_fd = creat(tee_path, 0600);
		
		if (tee_fd < 0) {
			ERROR("%s: %s\n", tee_path, strerror(errno));
			return -1;
		}
	}
	
	// pipe
	int fd[2];
	
	if (pipe(fd)) {
		ERROR("pipe(): %s\n", strerror(errno));
		return -1;
	}
	
	// fork
	pid_t child = fork();
	
	if (child == (pid_t)-1) {
		ERROR("fork(): %s\n", strerror(errno));
		return -1;
	}
	
	// parent
	if (child != 0) {
		// close tee file
		if (tee_fd >= 0)
			close(tee_fd);
		
		// close pipe's write end
		close(fd[1]);
		
		FILE *file = fdopen(fd[0], "r");
		
		if (!file) {
			ERROR("fdopen(pipefd): %s\n", strerror(errno));
			close(fd[0]);
			return -1;
		}
		
		p->child = child;
		p->file = file;
		return 0;
	}
	
	// child
	close(fd[0]);
	
	if (tee_fd >= 0) {
		// pipe
		int xfd[2];
		
		if (pipe(xfd)) {
			ERROR("pipe(): %s\n", strerror(errno));
			child_exit(EXIT_FAILURE);
		}
		
		// fork
		pid_t xchild = fork();
		
		if (xchild == (pid_t)-1) {
			ERROR("fork(): %s\n", strerror(errno));
			child_exit(EXIT_FAILURE);
		}
		
		// parent
		if (xchild != 0) {
			// close pipe's write end
			close(xfd[1]);
			
			char *buff = malloc(TEE_BUFFER);
			if (!buff) {
				ERROR("malloc() %s\n", strerror(errno));
				child_exit(EXIT_FAILURE);
			}
			
			while (1) {
				ssize_t n = read(xfd[0], buff, TEE_BUFFER);
				
				if (n < 0) {
					ERROR("read(): %s\n",
						strerror(errno));
					child_exit(EXIT_FAILURE);
				}
				
				if (n == 0)
					break;
				
				ssize_t w = write(fd[1], buff, n);
				
				if (w != n) {
					ERROR("write(%zd / %zd): %s\n",
						w, n, strerror(errno));
					child_exit(EXIT_FAILURE);
				}
				
				ssize_t t = write(tee_fd, buff, n);
				
				if (t != n) {
					ERROR("%s: %zd / %zd: %s\n",
						tee_path,
						t, n, strerror(errno));
					child_exit(EXIT_FAILURE);
				}
			}
			
			free(buff);
			close(xfd[0]);
			close(fd[1]);
			close(tee_fd);
			child_exit(EXIT_SUCCESS);
		}
		
		// child
		close(xfd[0]);
		close(fd[1]);
		close(tee_fd);
		
		fd[1] = xfd[1];
	}
	
	// exec
	if (fd_move(1, fd[1]))
		child_exit(EXIT_FAILURE);
	
	if (verbosity < 1) {
		if (fd_null(2))
			child_exit(EXIT_FAILURE);
	}
	
	execvp(argv[0], argv);
	
	ERROR("execvp('%s'): %s\n", argv[0], strerror(errno));
	child_exit(EXIT_FAILURE);
}



// ************************************************************************
// 
// ************************************************************************
int pipe_direct(struct piped_input *p, char *path)
{
	p->child = 0;
	p->file = fopen(path, "r");
	
	if (!p->file) {
		ERROR("%s: %s\n", path, strerror(errno));
		return -1;
	}
	
	return 0;
}

// ************************************************************************
// 
// ************************************************************************
int pipe_wait(struct piped_input *p)
{
	int r = (ferror(p->file)) ? EXIT_FAILURE : EXIT_SUCCESS;
	
	fclose(p->file);
	
	if (p->child != 0) {
		int status;
		
		pid_t rpid = waitpid(p->child, &status, 0);
		
		if ((size_t)rpid != p->child) {
			ERROR("waitpid(): %s\n", strerror(errno));
			r = EXIT_FAILURE;
		}
		
		if (!WIFEXITED(status))
			r = EXIT_FAILURE;
		
		if (r == EXIT_SUCCESS)
			r = WEXITSTATUS(status);
	}
	
	return r;
}

// ************************************************************************
// 
// ************************************************************************
